from abc import ABC, abstractmethod
import asyncio
import json
import pandapower as pp
import time
import math

import numpy as np 

# Strategy class from which all single requirements will inherit, to make this easily extendible 
class NeighbourhoodRequirementCheckStrategy(ABC):

    def __init__(self, border_regions, client_lms, vio_queue, logger):
        self.__br = border_regions
        self.client_lms = client_lms
        self.__vio_queue = vio_queue
        self.logger = logger
        self.__transformer_pos = -1
        
    @abstractmethod
    def check(self): 
        pass 
    
    # helper functions to get data
    async def get_data_from_lm(self, i_of_lm):
        """Get the latest data values from the specified local monitor."""
        data_node = None                  
        data_node = self.client_lms[i_of_lm]["data_node"]
        try:
            lm_data = await data_node.read_value()
        except Exception:
           lm_data = None
        
        big_data = []
        
        for meter in lm_data.__getattribute__("meters"):
            meterdata = meter 
    
            current = meterdata.__getattribute__("current")
            voltage = meterdata.__getattribute__("voltage")
            sensor_id = meterdata.__getattribute__("id")
            
            data = []
            data.append(sensor_id)
            data.append(current)
            data.append(voltage)
            
        
            big_data.append(data)
            
        return big_data
    
    async def get_data_from_sensor(self, name_of_sensor): 
        data_lm1 = await self.get_data_from_lm(0)
        data_lm2 = await self.get_data_from_lm(1)
        data_lm3 = await self.get_data_from_lm(2)
        data_lm4 = await self.get_data_from_lm(3)
        
        sensor_data = []
        
        for d in data_lm1: 
            if d[0] == name_of_sensor: 
                sensor_data = d
        
        for d in data_lm2: 
            if d[0] == name_of_sensor: 
               sensor_data = d

        for d in data_lm3: 
            if d[0] == name_of_sensor: 
               sensor_data = d

        for d in data_lm4: 
            if d[0] == name_of_sensor: 
               sensor_data = d
        
        return sensor_data
    
    async def get_v_data_from_sensor(self, name_of_sensor): 
        sensor_data = await self.get_data_from_sensor(name_of_sensor)
        if sensor_data:
           return sensor_data[2]
        else: 
            return []
    
    async def get_c_data_from_sensor(self, name_of_sensor): 
   
        sensor_data = await self.get_data_from_sensor(name_of_sensor)
        if sensor_data:
            return sensor_data[1]
        else: 
            return []
        
    async def get_lm_sm_data(self):
        data = []
        for i in range(3):
            sensors = await self.get_data_from_lm(i)
            for m in sensors: 
                if m[0] == "sensor_21":
                    data.append(m)
        
        return data
        
#REQ DEMKit case
class DEMKit_NM_Test_Case(NeighbourhoodRequirementCheckStrategy):
    async def check(self):
        """Checks Requirement S6: Only registered houses (or power generators) are feeding into the grid."""
        c1 = await self.get_c_data_from_sensor("sensor_21")
        lms = await self.get_data_from_lm(0)
        #print("Monitor1: " + str(lms[0][1]))
        lms = await self.get_data_from_lm(1)
        #print("Monitor2: " +  str(lms[1][1]))
        lms = await self.get_data_from_lm(2)
        print("Monitor3: " +  str(lms[2][1]))
        #lms = await self.get_data_from_lm(3)
        #print("Monitor4: " +  str(lms[3][1]))
    
#REQ DEMKit case
class DEMKit_S6_Registered_Feedin_Only(NeighbourhoodRequirementCheckStrategy):
    async def check(self):
        """Checks Requirement S6: Only registered houses (or power generators) are feeding into the grid."""        
        lm1_data = await self.get_data_from_lm(0)
        lm2_data = await self.get_data_from_lm(1)
        lm3_data = await self.get_data_from_lm(2)
        lm4_data = await self.get_data_from_lm(3)

        lm1_time = lm1_data[0][1]
        lm2_time = lm2_data[0][1]
        lm3_time = lm3_data[0][1]
        feeder_time = lm4_data[0][1]

        sum_sm_currents = float(lm1_data[1][1]) + float(lm2_data[1][1]) + float(lm3_data[1][1])
        feeder_current = float(lm4_data[1][1])
        power_flow_direction = (feeder_current >= 0)
        print(sum_sm_currents, feeder_current)
        # If neighborhood production < consumption:
        if power_flow_direction:
            # If total consumption of all combined houses exceeds the feeders measurement, an alert is thrown
            if (feeder_current + 10 < sum_sm_currents) and lm1_time == lm2_time == lm3_time == feeder_time:
                # Report to console
                self.logger.error("NM: Requirement S6 violated! Power flow: feeder(%sW) -> houses(%sW)", feeder_current, sum_sm_currents)
        # If neighborhood production < consumption:
        # If total production of all combined houses exceeds the feeders measurement, an alert is thrown
        else:
            if (feeder_current > sum_sm_currents + 10) and lm1_time == lm2_time == lm3_time == feeder_time:
                # Report to console
                self.logger.error("NM: Requirement S6 violated! Power flow: houses(%sW) -> feeder(%sW)", sum_sm_currents, feeder_current)

#REQ DEMKit case
# Looks at last 10 (smart meter) values of each local monitor and checks for sudden spikes
class DEMKit_S7_Anomaly_Power_Surge(NeighbourhoodRequirementCheckStrategy):
    last_value = [0, 0, 0]
    count = 0

    async def check(self):
        """Checks Requirement S7: Operator defined threshold of current in neighborhood is met for all meters."""
        smartmeter_data = await self.get_lm_sm_data()
        self.count += 1
        for i in range(3):
            value = smartmeter_data[i][1]
            last = self.last_value[i]
            if (abs(value) > (abs(last) * 7.5)) and (self.count > 3):
                self.logger.error("NM: Requirement S7 violated! The last recorded value for LM%s was %sW, but the most recent value is %sW!", (i + 1), last, value)
            self.last_value[i] = value